/*
 * Copyright (C) 2013 Castrense Savojardo
 * 	Bologna Biocomputing Group
 * 	University of Bologna, Italy
 *  savojard@biocomp.unibo.it
 *
 * align.c - This file is part of BCov
 *
 * BCov is a free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 3 of the License,
 * or (at your option) any later version.
 *
 * BCov is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BCov; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 */

#include <stdio.h>
#include <string.h>

#include "align.h"
#include "mem.h"

void copyaln(alignment * dest, alignment * source) {
	dest->size = source->size;
	dest->score = source->score;
	dest->pairs = allocvec(dest->size, sizeof(pair));
	memcpy(dest->pairs, source->pairs, source->size * sizeof(pair));
}

void backtrace_global(alignment * aln, entry ** H, unsigned int n, unsigned int m, unsigned int maxgaps) {
	int i, j, maxi, maxj, sizea;
	double max;
	pair * p;

	maxi = n;
	maxj = 1;
	max = H[maxi][maxj].score;
	for (j=2; j<=m; j++) {
		if (H[n][j].score >= max) {
			max = H[n][j].score;
			maxi = n;
			maxj = j;
		}
	}
	for (i=n-1; i>0; i--) {
		if (H[i][m].score > max) {
			max = H[i][m].score;
			maxi = i;
			maxj = m;
		}
	}

	sizea = 0;
	if (n <= m) {
		p = allocvec(n+maxgaps, sizeof(pair));
	} else {
		p = allocvec(m+maxgaps, sizeof(pair));
	}
	while (maxi > 0 && maxj > 0) {

		if (H[maxi][maxj].d == UP) {
			p[sizea].i = maxi - 1;
			p[sizea].j = GAP;
			maxi = maxi - 1;
		} else if (H[maxi][maxj].d == LEFT) {
			p[sizea].i = GAP;
			p[sizea].j = maxj - 1;
			maxj = maxj - 1;
		} else if (H[maxi][maxj].d == DIAG) {
			p[sizea].i = maxi - 1;
			p[sizea].j = maxj - 1;
			maxi = maxi - 1;
			maxj = maxj - 1;
		}
		sizea++;
	}
	aln->score = max;
	aln->pairs = allocvec(sizea, sizeof(pair));
	aln->size = sizea;
	j = 0;
	for (i = sizea - 1; i>=0; i--) {
		aln->pairs[j].i = p[i].i;
		aln->pairs[j].j = p[i].j;
		j++;
	}
	free(p);
}

void backtrace_local(alignment * aln, entry ** H, unsigned int n, unsigned int m, unsigned int maxgaps) {
	int i, j, maxi, maxj, sizea;
	double max;
	pair * p;


	maxi = 1;
	maxj = 1;
	max = H[maxi][maxj].score;
	for (i = 1; i <= n; i++) {
		for (j = 1; j <= m; j++) {
			if (H[i][j].score > max) {
				max = H[i][j].score;
				maxi = i;
				maxj = j;
			}
		}
	}

	sizea = 0;
	if (n <= m) {
		p = allocvec(n+maxgaps, sizeof(pair));
	} else {
		p = allocvec(m+maxgaps, sizeof(pair));
	}
	while (maxi > 0 && maxj > 0) {

		if (H[maxi][maxj].d == UP) {
			p[sizea].i = maxi - 1;
			p[sizea].j = GAP;
			maxi = maxi - 1;
		} else if (H[maxi][maxj].d == LEFT) {
			p[sizea].i = GAP;
			p[sizea].j = maxj - 1;
			maxj = maxj - 1;
		} else if (H[maxi][maxj].d == DIAG) {
			p[sizea].i = maxi - 1;
			p[sizea].j = maxj - 1;
			maxi = maxi - 1;
			maxj = maxj - 1;
		}
		sizea++;
	}

	aln->score = max;
	aln->pairs = allocvec(sizea, sizeof(pair));
	aln->size = sizea;
	j = 0;
	for (i = sizea - 1; i>=0; i--) {
		aln->pairs[j].i = p[i].i;
		aln->pairs[j].j = p[i].j;
		j++;
	}
	free(p);
}


void backtrace_maxov(alignment * aln, entry ** H, unsigned int n, unsigned int m) {
	int i, j, maxi, maxj, sizea, s, e;
	double max;
	pair * p;

	maxi = n;
	maxj = s = n <= m ? n - 1 : m - 1;
	max = H[maxi][maxj].score;
	for (j=s+1; j<=m; j++) {
		if (H[n][j].score >= max) {
			max = H[n][j].score;
			maxi = n;
			maxj = j;
		}
	}
	e = n <= m ? n - 2: m - 2;
	for (i=n-1; i>e; i--) {
		if (H[i][m].score > max) {
			max = H[i][m].score;
			maxi = i;
			maxj = m;
		}
	}
	sizea = 0;
	if (n <= m) {
		p = allocvec(n, sizeof(pair));
	} else {
		p = allocvec(m, sizeof(pair));
	}
	while (maxi > 0 && maxj > 0) {
		p[sizea].i = maxi - 1;
		p[sizea].j = maxj - 1;
		maxi = maxi - 1;
		maxj = maxj - 1;
		sizea++;
	}
	aln->score = max;
	aln->pairs = allocvec(sizea, sizeof(pair));
	aln->size = sizea;
	j = 0;
	for (i = sizea - 1; i>=0; i--) {
		aln->pairs[j].i = p[i].i;
		aln->pairs[j].j = p[i].j;
		j++;
	}
	free(p);
}

void align_strands(alignment * res, double ** S, unsigned int n, unsigned int m, unsigned int maxgaps, unsigned int type) {

	int i, j, maxi, maxj;
	double match, up, left;

	entry ** H;

	if (n == 1) {
		res->score = S[0][0];
		maxj = 0;
		for (j=1; j<m; j++) {
			if (S[0][j] > res->score) {
				res->score = S[0][j];
				maxj = j;
			}
		}
		res->size = 1;
		res->pairs = allocvec(1, sizeof(pair));
		res->pairs[0].i = 0;
		res->pairs[0].j = maxj;
	} else if (m == 1) {
		res->score = S[0][0];
		maxi = 0;
		for (i=1; i<n; i++) {
			if (S[i][0] > res->score) {
				res->score = S[i][0];
				maxi = i;
			}
		}
		res->size = 1;
		res->pairs = allocvec(1, sizeof(pair));
		res->pairs[0].i = maxi;
		res->pairs[0].j = 0;
	} else if (m > 1 && n > 1) {
		if (type == MAXOV) maxgaps = 0;
		H = (entry **) allocmat(n+1, m+1, sizeof(entry));
		for (i = 1; i < n+1; i++) {
			for (j = 1; j < m+1; j++) {
				match = H[i-1][j-1].score + S[i-1][j-1];
				up = H[i-1][j].gaps < maxgaps ? H[i-1][j].score : GAPP;
				left = H[i][j-1].gaps < maxgaps ? H[i][j-1].score : GAPP;

				if ((i == 1 && j == m) || (i == n && j == 1)) {
					up = H[i-1][j].matches > 0 ? up : GAPP;
					left = H[i][j-1].matches > 0 ? left : GAPP;
				}
				if (match >= up && match >= left) {
					H[i][j].score = match;
					H[i][j].matches = H[i][j-1].matches + 1;
					H[i][j].gaps = H[i-1][j-1].gaps;
					H[i][j].d = DIAG;
				} else if (up >= match && up >= left) {
					H[i][j].score = up;
					H[i][j].matches = H[i-1][j].matches;
					H[i][j].gaps = H[i-1][j].gaps + 1;
					H[i][j].d = UP;
				} else if (left > match && left > up) {
					H[i][j].score = left;
					H[i][j].matches = H[i][j-1].matches;
					H[i][j].gaps = H[i][j-1].gaps + 1;
					H[i][j].d = LEFT;
				}

			}
		}
		if (type == GLOBAL) {
			backtrace_global(res, H, n, m, maxgaps);
		} else if (type == LOCAL) {
			backtrace_local(res, H, n, m, maxgaps);
		} else if (type == MAXOV) {
			backtrace_maxov(res, H, n, m);
		}

		freemat(H);
	}
	//return res;
}
