/*
 * Copyright (C) 2013 Castrense Savojardo
 * 	Bologna Biocomputing Group
 * 	University of Bologna, Italy
 *  savojard@biocomp.unibo.it
 *
 * lp.c - This file is part of BCov
 *
 * BCov is a free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 3 of the License,
 * or (at your option) any later version.
 *
 * BCov is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BCov; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 */

#include <stdlib.h>
#include <glpk.h>

#include "lp.h"



double solve_problem(double ** D, double ** S, int ** C, unsigned int N) {

	int i, j, h, k;
	int * ia, * ja;
	double * ar;
	double score;

	glp_prob *mip = glp_create_prob();
	glp_set_prob_name(mip, "strand pairing");
	glp_set_obj_dir(mip, GLP_MAX);

	glp_add_rows(mip, N + N*(N-1)/2);
	for (i=1; i<=N; i++) {
		glp_set_row_bnds(mip, i, GLP_DB, 1.0, 2.0);
	}

	for (i=1; i<=N*(N-1)/2; i++) {
		glp_set_row_bnds(mip, N + i, GLP_FX, 0, 0);
	}

	glp_add_cols(mip, N * N);
	for (i=1; i<= N; i++) {
		for (j=1; j<= N; j++) {
			glp_set_col_kind(mip, (i-1)*N + j, GLP_BV);
			glp_set_obj_coef(mip, (i-1)*N + j, D[i-1][j-1]);
			if (C[i-1][j-1] == 0) {
				glp_set_col_bnds(mip, (i-1)*N + j, GLP_FX, 0, 0);
			}
		}
	}

	ia = calloc(1 + N * N + N * (N-1), sizeof(int));
	ja = calloc(1 + N * N + N * (N-1), sizeof(int));
	ar = calloc(1 + N * N + N * (N-1), sizeof(double));
	k=1;
	for (i=1; i<=N; i++) {
		for (j=1; j<=N; j++) {
			ia[k] = i;
			ja[k] = (i-1)*N + j;
			ar[k++] = 1;
		}
	}

    h = N+1;
	for (i=1; i<N; i++) {
		for (j=i+1; j<=N; j++) {
			ia[k] = h;
			ja[k] = (i-1)*N + j;
			ar[k++] = 1;
			ia[k] = h++;
			ja[k] = (j-1)*N + i;
			ar[k++] = -1;
		}
	}

	glp_load_matrix(mip, N * N + N * (N-1), ia, ja, ar);
	glp_iocp parm;
	glp_init_iocp(&parm);
	parm.presolve = GLP_ON;
	parm.msg_lev = VERBOSE;
	parm.tm_lim = 2 * 60 * 1000;
	glp_intopt(mip, &parm);
	score = glp_mip_obj_val(mip);

	for (i=1; i<=N; i++) {
		for (j=1; j<=N; j++) {
			S[i-1][j-1] = glp_mip_col_val(mip, (i-1)*N + j);
		}
	}
	glp_delete_prob(mip);
	free(ia);
	free(ja);
	free(ar);
	return score;
}
